/*
 * DEBUGGER code
 * 
 * On "publish", when buffer is free, debugger stores arbitrary variables 
 * content into, and mark this buffer as filled
 * 
 * 
 * Buffer content is read asynchronously, (from non real time part), 
 * and then buffer marked free again.
 *  
 * 
 * */
#ifdef TARGET_DEBUG_AND_RETAIN_DISABLE

void __init_debug    (void){}
void __cleanup_debug (void){}
void __retrieve_debug(void){}
void __publish_debug (void){}

#else

#include "iec_types_all.h"
#include "POUS.h"
/*for memcpy*/
#include <string.h>
#include <stdio.h>

#ifndef TARGET_ONLINE_DEBUG_DISABLE
#define BUFFER_SIZE 535

/* Atomically accessed variable for buffer state */
#define BUFFER_FREE 0
#define BUFFER_BUSY 1
static long buffer_state = BUFFER_FREE;

/* The buffer itself */
char debug_buffer[BUFFER_SIZE];

/* Buffer's cursor*/
static char* buffer_cursor = debug_buffer;
#endif

static unsigned int retain_offset = 0;
/***
 * Declare programs 
 **/
extern WATERSUPPLIER RES0__INSTANCE0;

/***
 * Declare global variables from resources and conf 
 **/
extern       WATERSUPPLIER   RES0__INSTANCE0;

typedef const struct {
    void *ptr;
    __IEC_types_enum type;
} dbgvardsc_t;

static dbgvardsc_t dbgvardsc[] = {
{&(RES0__INSTANCE0.CURRENTTIMEDT), LWORD_ENUM},
{&(RES0__INSTANCE0.CURRENTTIMETOD), DWORD_ENUM},
{&(RES0__INSTANCE0.SUPPLYTIME), DWORD_ENUM},
{&(RES0__INSTANCE0.NOOFSUPPLY), WORD_ENUM},
{&(RES0__INSTANCE0.TIMEMONITOR), WORD_ENUM},
{&(RES0__INSTANCE0.CURRENTNO), INT_ENUM},
{&(RES0__INSTANCE0.TESTSW), BOOL_ENUM},
{&(RES0__INSTANCE0.SVMONPL), BOOL_ENUM},
{&(RES0__INSTANCE0.READYTIMEPL), BOOL_ENUM},
{&(RES0__INSTANCE0.VALVRELAY1), BOOL_ENUM},
{&(RES0__INSTANCE0.VALVRELAY2), BOOL_ENUM},
{&(RES0__INSTANCE0.VALVRELAY3), BOOL_ENUM},
{&(RES0__INSTANCE0.READYTIME), BOOL_ENUM},
{&(RES0__INSTANCE0.READYTOCOMPARE), BOOL_ENUM},
{&(RES0__INSTANCE0.FLICK05_1), BOOL_ENUM},
{&(RES0__INSTANCE0.FLICK05_2), BOOL_ENUM},
{&(RES0__INSTANCE0.FLICK05), BOOL_ENUM},
{&(RES0__INSTANCE0.N1SUPPLY), BOOL_ENUM},
{&(RES0__INSTANCE0.N2SUPPLY), BOOL_ENUM},
{&(RES0__INSTANCE0.N3SUPPLY), BOOL_ENUM},
{&(RES0__INSTANCE0.N4SUPPLY), BOOL_ENUM},
{&(RES0__INSTANCE0.N5SUPPLY), BOOL_ENUM},
{&(RES0__INSTANCE0.N6SUPPLY), BOOL_ENUM},
{&(RES0__INSTANCE0.N7SUPPLY), BOOL_ENUM},
{&(RES0__INSTANCE0.N8SUPPLY), BOOL_ENUM},
{&(RES0__INSTANCE0.N9SUPPLY), BOOL_ENUM},
{&(RES0__INSTANCE0.SUPPLYTRIG), BOOL_ENUM},
{&(RES0__INSTANCE0.INSUPPLY), BOOL_ENUM},
{&(RES0__INSTANCE0.CNTEND), BOOL_ENUM},
{&(RES0__INSTANCE0.VALVEON), BOOL_ENUM},
{&(RES0__INSTANCE0.VALVEUP), BOOL_ENUM},
{&(RES0__INSTANCE0.CNTEND1), BOOL_ENUM},
{&(RES0__INSTANCE0.INTERVAL), BOOL_ENUM},
{&(RES0__INSTANCE0.INTERVALUP), BOOL_ENUM},
{&(RES0__INSTANCE0.TESTMODE), BOOL_ENUM},
{&(RES0__INSTANCE0.TESTTRIGER), BOOL_ENUM},
{&(RES0__INSTANCE0.NEXTINTERVAL1), BOOL_ENUM},
{&(RES0__INSTANCE0.NEXTTRIGER), BOOL_ENUM},
{&(RES0__INSTANCE0.DRVVALV0), BOOL_ENUM},
{&(RES0__INSTANCE0.DRVVALV1), BOOL_ENUM},
{&(RES0__INSTANCE0.DRVVALV2), BOOL_ENUM},
{&(RES0__INSTANCE0.TON0.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON0.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.TON0.IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON0.PT), TIME_ENUM},
{&(RES0__INSTANCE0.TON0.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.TON0.ET), TIME_ENUM},
{&(RES0__INSTANCE0.TON0.STATE), SINT_ENUM},
{&(RES0__INSTANCE0.TON0.PREV_IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON0.CURRENT_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.TON0.START_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.TON1.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON1.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.TON1.IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON1.PT), TIME_ENUM},
{&(RES0__INSTANCE0.TON1.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.TON1.ET), TIME_ENUM},
{&(RES0__INSTANCE0.TON1.STATE), SINT_ENUM},
{&(RES0__INSTANCE0.TON1.PREV_IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON1.CURRENT_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.TON1.START_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.TON2.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON2.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.TON2.IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON2.PT), TIME_ENUM},
{&(RES0__INSTANCE0.TON2.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.TON2.ET), TIME_ENUM},
{&(RES0__INSTANCE0.TON2.STATE), SINT_ENUM},
{&(RES0__INSTANCE0.TON2.PREV_IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON2.CURRENT_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.TON2.START_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.CTD0.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.CTD0.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.CTD0.CD), BOOL_ENUM},
{&(RES0__INSTANCE0.CTD0.LD), BOOL_ENUM},
{&(RES0__INSTANCE0.CTD0.PV), INT_ENUM},
{&(RES0__INSTANCE0.CTD0.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.CTD0.CV), INT_ENUM},
{&(RES0__INSTANCE0.CTD0.CD_T.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.CTD0.CD_T.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.CTD0.CD_T.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.CTD0.CD_T.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.CTD0.CD_T.M), BOOL_ENUM},
{&(RES0__INSTANCE0.TON3.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON3.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.TON3.IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON3.PT), TIME_ENUM},
{&(RES0__INSTANCE0.TON3.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.TON3.ET), TIME_ENUM},
{&(RES0__INSTANCE0.TON3.STATE), SINT_ENUM},
{&(RES0__INSTANCE0.TON3.PREV_IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON3.CURRENT_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.TON3.START_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.TON4.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON4.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.TON4.IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON4.PT), TIME_ENUM},
{&(RES0__INSTANCE0.TON4.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.TON4.ET), TIME_ENUM},
{&(RES0__INSTANCE0.TON4.STATE), SINT_ENUM},
{&(RES0__INSTANCE0.TON4.PREV_IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON4.CURRENT_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.TON4.START_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.TON5.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON5.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.TON5.IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON5.PT), TIME_ENUM},
{&(RES0__INSTANCE0.TON5.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.TON5.ET), TIME_ENUM},
{&(RES0__INSTANCE0.TON5.STATE), SINT_ENUM},
{&(RES0__INSTANCE0.TON5.PREV_IN), BOOL_ENUM},
{&(RES0__INSTANCE0.TON5.CURRENT_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.TON5.START_TIME), TIME_ENUM},
{&(RES0__INSTANCE0.R_TRIG1.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG1.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG1.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG1.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG1.M), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE237_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE237_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.LWORD_TO_DT8_OUT), DT_ENUM},
{&(RES0__INSTANCE0.DATE_AND_TIME_TO_TIME_OF_DAY9_OUT), TOD_ENUM},
{&(RES0__INSTANCE0.TOD_TO_DWORD10_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.DWORD_TO_WORD273_OUT), WORD_ENUM},
{&(RES0__INSTANCE0.GT37_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.GT37_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG2.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG2.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG2.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG2.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG2.M), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE52_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE52_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.GT40_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.GT40_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG3.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG3.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG3.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG3.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG3.M), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE64_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE64_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.GT62_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.GT62_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG4.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG4.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG4.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG4.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG4.M), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE75_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE75_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.GT81_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.GT81_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG5.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG5.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG5.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG5.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG5.M), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE89_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE89_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.GT95_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.GT95_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG6.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG6.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG6.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG6.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG6.M), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE103_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE103_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.GT109_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.GT109_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG7.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG7.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG7.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG7.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG7.M), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE117_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE117_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.GT123_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.GT123_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG8.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG8.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG8.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG8.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG8.M), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE131_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE131_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.GT137_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.GT137_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG9.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG9.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG9.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG9.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG9.M), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE145_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE145_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.GT151_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.GT151_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG10.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG10.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG10.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG10.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG10.M), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE159_ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.MOVE159_OUT), DWORD_ENUM},
{&(RES0__INSTANCE0.R_TRIG11.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG11.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG11.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG11.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG11.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG12.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG12.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG12.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG12.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG12.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG13.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG13.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG13.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG13.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG13.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG14.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG14.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG14.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG14.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG14.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG15.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG15.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG15.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG15.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG15.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG16.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG16.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG16.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG16.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG16.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG17.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG17.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG17.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG17.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG17.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG18.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG18.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG18.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG18.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG18.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG19.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG19.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG19.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG19.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG19.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG20.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG20.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG20.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG20.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG20.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG21.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG21.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG21.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG21.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG21.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG22.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG22.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG22.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG22.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG22.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG23.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG23.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG23.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG23.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG23.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG24.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG24.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG24.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG24.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG24.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG25.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG25.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG25.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG25.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG25.M), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG26.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG26.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG26.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG26.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG26.M), BOOL_ENUM},
{&(RES0__INSTANCE0.DWORD_TO_TIME202_OUT), TIME_ENUM},
{&(RES0__INSTANCE0.R_TRIG27.EN), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG27.ENO), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG27.CLK), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG27.Q), BOOL_ENUM},
{&(RES0__INSTANCE0.R_TRIG27.M), BOOL_ENUM},
{&(RES0__INSTANCE0.EQ288_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.EQ294_OUT), BOOL_ENUM},
{&(RES0__INSTANCE0.EQ305_OUT), BOOL_ENUM}
};

typedef void(*__for_each_variable_do_fp)(dbgvardsc_t*);
void __for_each_variable_do(__for_each_variable_do_fp fp)
{
    unsigned int i;
    for(i = 0; i < sizeof(dbgvardsc)/sizeof(dbgvardsc_t); i++){
        dbgvardsc_t *dsc = &dbgvardsc[i];
        if(dsc->type != UNKNOWN_ENUM) 
            (*fp)(dsc);
    }
}

#define __Unpack_case_t(TYPENAME) \
        case TYPENAME##_ENUM :\
            *flags = ((__IEC_##TYPENAME##_t *)varp)->flags;\
            forced_value_p = *real_value_p = &((__IEC_##TYPENAME##_t *)varp)->value;\
            break;

#define __Unpack_case_p(TYPENAME)\
        case TYPENAME##_O_ENUM :\
            *flags = __IEC_OUTPUT_FLAG;\
        case TYPENAME##_P_ENUM :\
            *flags |= ((__IEC_##TYPENAME##_p *)varp)->flags;\
            *real_value_p = ((__IEC_##TYPENAME##_p *)varp)->value;\
            forced_value_p = &((__IEC_##TYPENAME##_p *)varp)->fvalue;\
            break;

void* UnpackVar(dbgvardsc_t *dsc, void **real_value_p, char *flags)
{
    void *varp = dsc->ptr;
    void *forced_value_p = NULL;
    *flags = 0;
    /* find data to copy*/
    switch(dsc->type){
        __ANY(__Unpack_case_t)
        __ANY(__Unpack_case_p)
    default:
        break;
    }
    if (*flags & __IEC_FORCE_FLAG)
        return forced_value_p;
    return *real_value_p;
}

void Remind(unsigned int offset, unsigned int count, void * p);

void RemindIterator(dbgvardsc_t *dsc)
{
    void *real_value_p = NULL;
    char flags = 0;
    UnpackVar(dsc, &real_value_p, &flags);

    if(flags & __IEC_RETAIN_FLAG){
        USINT size = __get_type_enum_size(dsc->type);
        /* compute next cursor positon*/
        unsigned int next_retain_offset = retain_offset + size;
        /* if buffer not full */
        Remind(retain_offset, size, real_value_p);
        /* increment cursor according size*/
        retain_offset = next_retain_offset;
    }
}

extern int CheckRetainBuffer(void);
extern void InitRetain(void);

void __init_debug(void)
{
    /* init local static vars */
#ifndef TARGET_ONLINE_DEBUG_DISABLE	
    buffer_cursor = debug_buffer;
    buffer_state = BUFFER_FREE;
#endif

    retain_offset = 0;
    InitRetain();
    /* Iterate over all variables to fill debug buffer */
    if(CheckRetainBuffer()){
    	__for_each_variable_do(RemindIterator);
    }else{
    	char mstr[] = "RETAIN memory invalid - defaults used";
        LogMessage(LOG_WARNING, mstr, sizeof(mstr));
    }
    retain_offset = 0;
}

extern void InitiateDebugTransfer(void);
extern void CleanupRetain(void);

extern unsigned long __tick;

void __cleanup_debug(void)
{
#ifndef TARGET_ONLINE_DEBUG_DISABLE	
    buffer_cursor = debug_buffer;
    InitiateDebugTransfer();
#endif    

    CleanupRetain();
}

void __retrieve_debug(void)
{
}


void Retain(unsigned int offset, unsigned int count, void * p);

static inline void BufferIterator(dbgvardsc_t *dsc, int do_debug)
{
    void *real_value_p = NULL;
    void *visible_value_p = NULL;
    char flags = 0;

    visible_value_p = UnpackVar(dsc, &real_value_p, &flags);

    if(flags & ( __IEC_DEBUG_FLAG | __IEC_RETAIN_FLAG)){
        USINT size = __get_type_enum_size(dsc->type);

#ifndef TARGET_ONLINE_DEBUG_DISABLE	
        if(flags & __IEC_DEBUG_FLAG){
            /* copy visible variable to buffer */;
            if(do_debug){
                /* compute next cursor positon.
                   No need to check overflow, as BUFFER_SIZE
                   is computed large enough */
		if((dsc->type == STRING_ENUM)   ||
		   (dsc->type == STRING_P_ENUM) ||
		   (dsc->type == STRING_O_ENUM)){
                    /* optimization for strings */
                    size = ((STRING*)visible_value_p)->len + 1;
                }
                char* next_cursor = buffer_cursor + size;
                /* copy data to the buffer */
                memcpy(buffer_cursor, visible_value_p, size);
                /* increment cursor according size*/
                buffer_cursor = next_cursor;
            }
            /* re-force real value of outputs (M and Q)*/
            if((flags & __IEC_FORCE_FLAG) && (flags & __IEC_OUTPUT_FLAG)){
                memcpy(real_value_p, visible_value_p, size);
            }
        }
#endif	

        if(flags & __IEC_RETAIN_FLAG){
            /* compute next cursor positon*/
            unsigned int next_retain_offset = retain_offset + size;
            /* if buffer not full */
            Retain(retain_offset, size, real_value_p);
            /* increment cursor according size*/
            retain_offset = next_retain_offset;
        }
    }
}

void DebugIterator(dbgvardsc_t *dsc){
    BufferIterator(dsc, 1);
}

void RetainIterator(dbgvardsc_t *dsc){
    BufferIterator(dsc, 0);
}


unsigned int retain_size = 0;

/* GetRetainSizeIterator */
void GetRetainSizeIterator(dbgvardsc_t *dsc)
{
    void *real_value_p = NULL;
    char flags = 0;
    UnpackVar(dsc, &real_value_p, &flags);

    if(flags & __IEC_RETAIN_FLAG){
        USINT size = __get_type_enum_size(dsc->type);
        /* Calc retain buffer size */
        retain_size += size;
    }
}

/* Return size of all retain variables */
unsigned int GetRetainSize(void)
{
    __for_each_variable_do(GetRetainSizeIterator);
    return retain_size;
}


extern void PLC_GetTime(IEC_TIME*);
extern int TryEnterDebugSection(void);
extern long AtomicCompareExchange(long*, long, long);
extern long long AtomicCompareExchange64(long long* , long long , long long);
extern void LeaveDebugSection(void);
extern void ValidateRetainBuffer(void);
extern void InValidateRetainBuffer(void);

void __publish_debug(void)
{
    retain_offset = 0;
    InValidateRetainBuffer();
    
#ifndef TARGET_ONLINE_DEBUG_DISABLE 
    /* Check there is no running debugger re-configuration */
    if(TryEnterDebugSection()){
        /* Lock buffer */
        long latest_state = AtomicCompareExchange(
            &buffer_state,
            BUFFER_FREE,
            BUFFER_BUSY);
            
        /* If buffer was free */
        if(latest_state == BUFFER_FREE)
        {
            /* Reset buffer cursor */
            buffer_cursor = debug_buffer;
            /* Iterate over all variables to fill debug buffer */
            __for_each_variable_do(DebugIterator);
            
            /* Leave debug section,
             * Trigger asynchronous transmission 
             * (returns immediately) */
            InitiateDebugTransfer(); /* size */
        }else{
            /* when not debugging, do only retain */
            __for_each_variable_do(RetainIterator);
        }
        LeaveDebugSection();
    }else
#endif
    {
        /* when not debugging, do only retain */
        __for_each_variable_do(RetainIterator);
    }
    ValidateRetainBuffer();
}

#ifndef TARGET_ONLINE_DEBUG_DISABLE
#define __RegisterDebugVariable_case_t(TYPENAME) \
        case TYPENAME##_ENUM :\
            ((__IEC_##TYPENAME##_t *)varp)->flags |= flags;\
            if(force)\
             ((__IEC_##TYPENAME##_t *)varp)->value = *((TYPENAME *)force);\
            break;
#define __RegisterDebugVariable_case_p(TYPENAME)\
        case TYPENAME##_P_ENUM :\
            ((__IEC_##TYPENAME##_p *)varp)->flags |= flags;\
            if(force)\
             ((__IEC_##TYPENAME##_p *)varp)->fvalue = *((TYPENAME *)force);\
            break;\
        case TYPENAME##_O_ENUM :\
            ((__IEC_##TYPENAME##_p *)varp)->flags |= flags;\
            if(force){\
             ((__IEC_##TYPENAME##_p *)varp)->fvalue = *((TYPENAME *)force);\
             *(((__IEC_##TYPENAME##_p *)varp)->value) = *((TYPENAME *)force);\
            }\
            break;
void RegisterDebugVariable(unsigned int idx, void* force)
{
    if(idx  < sizeof(dbgvardsc)/sizeof(dbgvardsc_t)){
        unsigned char flags = force ?
            __IEC_DEBUG_FLAG | __IEC_FORCE_FLAG :
            __IEC_DEBUG_FLAG;
        dbgvardsc_t *dsc = &dbgvardsc[idx];
        void *varp = dsc->ptr;
        switch(dsc->type){
            __ANY(__RegisterDebugVariable_case_t)
            __ANY(__RegisterDebugVariable_case_p)
        default:
            break;
        }
    }
}

#define __ResetDebugVariablesIterator_case_t(TYPENAME) \
        case TYPENAME##_ENUM :\
            ((__IEC_##TYPENAME##_t *)varp)->flags &= ~(__IEC_DEBUG_FLAG|__IEC_FORCE_FLAG);\
            break;

#define __ResetDebugVariablesIterator_case_p(TYPENAME)\
        case TYPENAME##_P_ENUM :\
        case TYPENAME##_O_ENUM :\
            ((__IEC_##TYPENAME##_p *)varp)->flags &= ~(__IEC_DEBUG_FLAG|__IEC_FORCE_FLAG);\
            break;

void ResetDebugVariablesIterator(dbgvardsc_t *dsc)
{
    /* force debug flag to 0*/
    void *varp = dsc->ptr;
    switch(dsc->type){
        __ANY(__ResetDebugVariablesIterator_case_t)
        __ANY(__ResetDebugVariablesIterator_case_p)
    default:
        break;
    }
}

void ResetDebugVariables(void)
{
    __for_each_variable_do(ResetDebugVariablesIterator);
}

void FreeDebugData(void)
{
    /* atomically mark buffer as free */
    AtomicCompareExchange(
        &buffer_state,
        BUFFER_BUSY,
        BUFFER_FREE);
}
int WaitDebugData(unsigned long *tick);
/* Wait until debug data ready and return pointer to it */
int GetDebugData(unsigned long *tick, unsigned long *size, void **buffer){
    int wait_error = WaitDebugData(tick);
    if(!wait_error){
        *size = buffer_cursor - debug_buffer;
        *buffer = debug_buffer;
    }
    return wait_error;
}
#endif
#endif

